@@ -55,33 +55,52 @@ class BadValue(Exception):
 
 # Statement elements (see RFC, section 8.3)
 # They are used in different commands.
-comparator = {"name" : "comparator",
-              "type" : ["tag"],
-              "values" : [":comparator"],
-              "extra_arg" : {"type" : "string",
-                             "values" : ['"i;octet"', '"i;ascii-casemap"']},
-              "required" : False}
-address_part = {"name" : "address-part",
-                "values" : [":localpart", ":domain", ":all"],
-                "type" : ["tag"],
-                "required" : False}
-match_type = {"name" : "match-type",
-              "values" : [":is", ":contains", ":matches"],
-              "type" : ["tag"],
-              "required" : False}
+comparator = {
+        "name": "comparator",
+        "type": [ "tag" ],
+        "values": [ ":comparator" ],
+        "extra_arg": {
+                "type" : "string",
+                "values" : [ '"i;octet"', '"i;ascii-casemap"' ],
+            },
+        "required": False
+    }
+
+address_part = {
+        "name": "address-part",
+        "values": [ ":localpart", ":domain", ":all" ],
+        "type" : [ "tag" ],
+        "required" : False
+    }
+
+match_type = {
+        "name": "match-type",
+        "values": [ ":is", ":contains", ":matches" ],
+        "type": [ "tag" ],
+        "required": False
+    }
+
+namespace = {
+        "name": "namespace",
+        "values": [ ":personal", ":global" ],
+        "type": [ "tag" ],
+        "required": False
+    }
 
 class Command(object):
-    """Generic command representation.
+    """
+        Generic command representation.
+
+        A command is described as follow:
+
+            * A name
+            * A type
+            * A description of supported arguments
+            * Does it accept an unkwown quantity of arguments? (ex: anyof, allof)
+            * Does it accept children? (ie. subcommands)
+            * Is it an extension?
+            * Must follow only certain commands
 
-    A command is described as follow:
-     * A name
-     * A type
-     * A description of supported arguments
-     * Does it accept an unkwown quantity of arguments? (ex: anyof, allof)
-     * Does it accept children? (ie. subcommands)
-     * Is it an extension?
-     * Must follow only certain commands
-    
     """
     _type = None
     variable_args_nb = False
@@ -97,7 +116,9 @@ class Command(object):
         self.nextargpos = 0
         self.required_args = -1
         self.rargs_cnt = 0
-        self.curarg = None # for arguments that expect an argument :p (ex: :comparator)
+
+        # for arguments that expect an argument :p (ex: :comparator)
+        self.curarg = None
 
         self.name = self.__class__.__name__.replace("Command", "")
         self.name = self.name.lower()
@@ -106,12 +127,13 @@ class Command(object):
         return "%s (type: %s)" % (self.name, self._type)
 
     def tosieve(self, indentlevel=0, target=sys.stdout):
-        """Generate the sieve syntax corresponding to this command
+        """
+            Generate the sieve syntax corresponding to this command
 
-        Recursive method.
+            Recursive method.
 
-        :param indentlevel: current indentation level
-        :param target: opened file pointer where the content will be printed
+            :param indentlevel: current indentation level
+            :param target: opened file pointer where the content will be printed
         """
         self.__print(self.name, indentlevel, nocr=True, target=target)
         if self.has_arguments():
@@ -143,18 +165,24 @@ class Command(object):
 
                 if "string" in arg["type"]:
                     target.write('"%s"' % value.strip('"'))
+                elif arg.has_key('extra_arg') and "string" in arg['extra_arg']['type']:
+                    target.write('"%s"' % value.strip('"'))
                 else:
-                    target.write(value)
+                    target.write("%s" % value)
 
         if not self.accept_children:
             if self.get_type() != "test":
                 print >>target, ";"
             return
+
         if self.get_type() != "control":
             return
+
         print >>target, " {"
+
         for ch in self.children:
             ch.tosieve(indentlevel + 4, target=target)
+
         self.__print("}", indentlevel, target=target)
 
     def __print(self, data, indentlevel, nocr=False, target=sys.stdout):
@@ -165,37 +193,46 @@ class Command(object):
             print >>target, text
 
     def __get_arg_type(self, arg):
-        """Return the type corresponding to the given name.
-
-        :param arg: a defined argument name
         """
+            Return the type corresponding to the given name.
+
+            :param arg: a defined argument name
+        """
+
         for a in self.args_definition:
             if a["name"] == arg:
                 return a["type"]
         return None
 
     def complete_cb(self):
-        """Completion callback
-        
-        Called when a command is considered as complete by the parser.
         """
+            Completion callback
+
+            Called when a command is considered as complete by the parser.
+        """
+
         pass
 
     def get_expected_first(self):
-        """Return the first expected token for this command"""
+        """
+            Return the first expected token for this command
+        """
+
         return None
 
     def has_arguments(self):
         return len(self.args_definition) != 0
 
     def dump(self, indentlevel=0, target=sys.stdout):
-        """Display the command
-
-        Pretty printing of this command and its eventual arguments and
-        children. (recursively)
-
-        :param indentlevel: integer that indicates indentation level to apply
         """
+            Display the command
+
+            Pretty printing of this command and its eventual arguments and
+            children. (recursively)
+
+            :param indentlevel: integer that indicates indentation level to apply
+        """
+
         self.__print(self, indentlevel, target=target)
         indentlevel += 4
         if self.has_arguments():
@@ -218,28 +255,32 @@ class Command(object):
             ch.dump(indentlevel, target)
 
     def addchild(self, child):
-        """Add a new child to the command
-
-        A child corresponds to a command located into a block (this
-        command's block). It can be either an action or a control.
-
-        :param child: the new child
-        :return: True on succes, False otherwise
         """
+            Add a new child to the command
+
+            A child corresponds to a command located into a block (this
+            command's block). It can be either an action or a control.
+
+            :param child: the new child
+            :return: True on succes, False otherwise
+        """
+
         if not self.accept_children:
             return False
         self.children += [child]
         return True
 
     def iscomplete(self):
-        """Check if the command is complete
-
-        Check if all required arguments have been encountered. For
-        commands that allow an undefined number of arguments, this
-        method always returns False.
-
-        :return: True if command is complete, False otherwise
         """
+            Check if the command is complete
+
+            Check if all required arguments have been encountered. For
+            commands that allow an undefined number of arguments, this
+            method always returns False.
+
+            :return: True if command is complete, False otherwise
+        """
+
         if self.variable_args_nb:
             return False
         if self.required_args == -1:
@@ -250,54 +291,69 @@ class Command(object):
         return self.rargs_cnt == self.required_args
 
     def get_type(self):
-        """Return the command's type"""
+        """
+            Return the command's type
+        """
+
         if self._type is None:
             raise NotImplemented
         return self._type
 
     def __is_valid_value_for_arg(self, arg, value):
-        """Check if value is allowed for arg
-
-        Some commands only allow a limited set of values. The method
-        always returns True for methods that do not provide such a
-        set.
-
-        :param arg: the argument's name
-        :param value: the value to check
-        :return: True on succes, False otherwise
         """
+            Check if value is allowed for arg
+
+            Some commands only allow a limited set of values. The method
+            always returns True for methods that do not provide such a
+            set.
+
+            :param arg: the argument's name
+            :param value: the value to check
+            :return: True on succes, False otherwise
+        """
+
         if not arg.has_key("values"):
             return True
+
+        if not isinstance(value, str) and not isinstance(value, basestring):
+            if arg.has_key('type') and "stringlist" in arg['type']:
+                return True
+            else:
+                return False
+
         return value.lower() in arg["values"]
 
     def check_next_arg(self, atype, avalue, add=True):
-        """Argument validity checking
-
-        This method is usually used by the parser to check if detected
-        argument is allowed for this command.
-
-        We make a distinction between required and optional
-        arguments. Optional (or tagged) arguments can be provided
-        unordered but the required ones.
-
-        A special handling is also for arguments that require an
-        argument (example: the :comparator argument expects a string
-        argument).
-
-        The "testlist" type is checked separately as we can't know in
-        advance how many arguments will be provided.
-
-        If the argument is incorrect, the method raises the
-        appropriate exception, or return False to let the parser
-        handle the exception.
-
-        :param atype: the argument's type
-        :param avalue: the argument's value
-        :param add: indicates if this argument should be recorded on success
-        :return: True on success, False otherwise
         """
+            Argument validity checking
+
+            This method is usually used by the parser to check if detected
+            argument is allowed for this command.
+
+            We make a distinction between required and optional
+            arguments. Optional (or tagged) arguments can be provided
+            unordered but the required ones.
+
+            A special handling is also for arguments that require an
+            argument (example: the :comparator argument expects a string
+            argument).
+
+            The "testlist" type is checked separately as we can't know in
+            advance how many arguments will be provided.
+
+            If the argument is incorrect, the method raises the
+            appropriate exception, or return False to let the parser
+            handle the exception.
+
+            :param atype: the argument's type
+            :param avalue: the argument's value
+            :param add: indicates if this argument should be recorded on success
+            :return: True on success, False otherwise
+        """
+
         if not self.has_arguments():
             return False
+
         if self.iscomplete():
             return False
 
@@ -307,53 +363,57 @@ class Command(object):
                         avalue in self.curarg["extra_arg"]["values"]:
                     if add:
                         self.arguments[self.curarg["name"]] = avalue
+                    self.nextargpos += 1
                     self.curarg = None
                     return True
+
             raise BadValue(self.curarg["name"], avalue)
 
         failed = False
-        pos = self.nextargpos
-        while pos < len(self.args_definition):
-            curarg = self.args_definition[pos]
-            if curarg["required"]:
-                if curarg["type"] == ["testlist"]:
-                    if atype != "test":
+
+        for arg in self.args_definition[self.nextargpos:]:
+            if self.arguments.has_key(arg["name"]):
+                continue
+
+            if arg["required"]:
+                if arg["type"] == ["testlist"]:
+                    if not atype == "test":
                         failed = True
                     elif add:
-                        if not self.arguments.has_key(curarg["name"]):
-                            self.arguments[curarg["name"]] = []
-                        self.arguments[curarg["name"]] += [avalue]
-                elif atype not in curarg["type"] or \
-                        not self.__is_valid_value_for_arg(curarg, avalue):
+                        if not self.arguments.has_key(arg["name"]):
+                            self.arguments[arg["name"]] = []
+                        self.arguments[arg["name"]].append(avalue)
+                elif atype not in arg["type"] or \
+                    not self.__is_valid_value_for_arg(arg, avalue):
+
                     failed = True
                 else:
-                    self.rargs_cnt += 1
-                    self.nextargpos = pos + 1
+                    self.nextargpos += 1
                     if add:
-                        self.arguments[curarg["name"]] = avalue
+                        self.arguments[arg["name"]] = avalue
+
                 break
 
-            if atype in curarg["type"]:
-                if self.__is_valid_value_for_arg(curarg, avalue):
-                    if curarg.has_key("extra_arg"):
-                        self.curarg = curarg
-                        break
-                    if add:
-                        self.arguments[curarg["name"]] = avalue
+            if self.__is_valid_value_for_arg(arg, avalue):
+                self.arguments[arg["name"]] = avalue
+                if arg.has_key('extra_arg'):
+                    self.curarg = arg
                     break
-
-            pos += 1
+                self.nextargpos += 1
+                break
 
         if failed:
-            raise BadArgument(self.name, avalue, 
-                              self.args_definition[pos]["type"])
+            raise BadArgument(self.name, avalue,
+                              self.args_definition[self.nextargpos]["type"])
         return True
 
     def __getitem__(self, name):
-        """Shorcut to access a command argument
-
-        :param name: the argument's name
         """
+            Shorcut to access a command argument
+
+            :param name: the argument's name
+        """
+
         found = False
         for ad in self.args_definition:
             if ad["name"] == name:
@@ -366,20 +426,27 @@ class Command(object):
         return self.arguments[name]
 
 class ControlCommand(Command):
-    """Indermediate class to represent "control" commands"""
-    _type = "control"
-    
-class RequireCommand(ControlCommand):
-    """The 'require' command
-
-    This class has one big difference with others as it is used to
-    store loaded extension names. (The result is we can check for
-    unloaded extensions during the parsing)
     """
+        Indermediate class to represent "control" commands
+    """
+
+    _type = "control"
+
+class RequireCommand(ControlCommand):
+    """
+        The 'require' command
+
+        This class has one big difference with others as it is used to
+        store loaded extension names. (The result is we can check for
+        unloaded extensions during the parsing)
+    """
+
     args_definition = [
-        {"name" : "capabilities",
-         "type" : ["string", "stringlist"],
-         "required" : True}
+                {
+                    "name": "capabilities",
+                    "type": [ "string", "stringlist" ],
+                    "required": True
+            }
         ]
 
     loaded_extensions = []
@@ -401,9 +468,11 @@ class IfCommand(ControlCommand):
     accept_children = True
 
     args_definition = [
-        {"name" : "test",
-         "type" : ["test"],
-         "required" : True}
+            {
+                    "name": "test",
+                    "type": [ "test" ],
+                    "required": True
+                }
         ]
 
     def get_expected_first(self):
@@ -413,9 +482,9 @@ class ElsifCommand(ControlCommand):
     accept_children = True
     must_follow = ["if", "elsif"]
     args_definition = [
-        {"name" : "test",
-         "type" : ["test"],
-         "required" : True}
+        {"name": "test",
+         "type": [ "test" ],
+         "required": True}
         ]
 
     def get_expected_first(self):
@@ -433,16 +502,53 @@ class ActionCommand(Command):
 class FileintoCommand(ActionCommand):
     is_extension = True
     args_definition = [
-        {"name" : "mailbox",
-         "type" : ["string"],
-         "required" : True}
+        {"name": "mailbox",
+         "type": [ "string" ],
+         "required": True}
+        ]
+
+class IncludeCommand(ControlCommand):
+    is_extension = True
+    accept_children = False
+    args_definition = [
+        namespace,
+        {"name": "script",
+         "type": [ "string" ],
+         "required": True}
+        ]
+
+class AddflagCommand(ActionCommand):
+    args_definition = [
+            {
+                    "name": "flag",
+                    "type": [ "string" ],
+                    "required": True
+                }
+        ]
+
+class SetflagCommand(ActionCommand):
+    args_definition = [
+            {
+                    "name": "flag",
+                    "type": [ "string" ],
+                    "required": True
+                }
         ]
 
 class RedirectCommand(ActionCommand):
+    is_extension = False
     args_definition = [
-        {"name" : "address",
-         "type" : ["string"],
-         "required" : True}
+            {
+                    "name": "copy",
+                    "values": [ ":copy" ],
+                    "type": [ "tag" ],
+                    "required": False
+                },
+            {
+                    "name": "address",
+                    "type": [ "string", "stringlist" ],
+                    "required": True
+                },
         ]
 
 class KeepCommand(ActionCommand):
@@ -460,12 +566,12 @@ class AddressCommand(TestCommand):
         comparator,
         address_part,
         match_type,
-        {"name" : "header-list",
-         "type" : ["string", "stringlist"],
-         "required" : True},
-        {"name" : "key-list",
-         "type" : ["string", "stringlist"],
-         "required" : True}
+        {"name": "header-list",
+         "type": [ "string", "stringlist" ],
+         "required": True},
+        {"name": "key-list",
+         "type": [ "string", "stringlist" ],
+         "required": True}
         ]
 
 class AllofCommand(TestCommand):
@@ -473,9 +579,9 @@ class AllofCommand(TestCommand):
     variable_args_nb = True
 
     args_definition = [
-        {"name" : "tests",
-         "type" : ["testlist"],
-         "required" : True}
+        {"name": "tests",
+         "type": [ "testlist" ],
+         "required": True}
         ]
 
     def get_expected_first(self):
@@ -486,9 +592,9 @@ class AnyofCommand(TestCommand):
     variable_args_nb = True
 
     args_definition = [
-        {"name" : "tests",
-         "type" : ["testlist"],
-         "required" : True}
+        {"name": "tests",
+         "type": [ "testlist" ],
+         "required": True}
         ]
 
     def get_expected_first(self):
@@ -499,19 +605,19 @@ class EnvelopeCommand(TestCommand):
         comparator,
         address_part,
         match_type,
-        {"name" : "header-list",
-         "type" : ["string", "stringlist"],
-         "required" : True},
-        {"name" : "key-list",
-         "type" : ["string", "stringlist"],
-         "required" : True}
+        {"name": "header-list",
+         "type": [ "string", "stringlist" ],
+         "required": True},
+        {"name": "key-list",
+         "type": [ "string", "stringlist" ],
+         "required": True}
         ]
 
 class ExistsCommand(TestCommand):
     args_definition = [
-        {"name" : "header-names",
-         "type" : ["stringlist"],
-         "required" : True}
+        {"name": "header-names",
+         "type": [ "string", "stringlist" ],
+         "required": True}
         ]
 
 class TrueCommand(TestCommand):
@@ -524,21 +630,21 @@ class HeaderCommand(TestCommand):
     args_definition = [
         comparator,
         match_type,
-        {"name" : "header-names",
-         "type" : ["string", "stringlist"],
-         "required" : True},
-        {"name" : "key-list",
-         "type" : ["string", "stringlist"],
-         "required" : True}
+        {"name": "header-names",
+         "type": [ "string", "stringlist" ],
+         "required": True},
+        {"name": "key-list",
+         "type": [ "string", "stringlist" ],
+         "required": True}
         ]
 
 class NotCommand(TestCommand):
     accept_children = True
 
     args_definition = [
-        {"name" : "test",
-         "type" : ["test"],
-         "required" : True}
+        {"name": "test",
+         "type": [ "test" ],
+         "required": True}
         ]
 
     def get_expected_first(self):
@@ -546,55 +652,55 @@ class NotCommand(TestCommand):
 
 class SizeCommand(TestCommand):
     args_definition = [
-        {"name" : "comparator",
-         "type" : ["tag"],
-         "values" : [":over", ":under"],
-         "required" : True},
-        {"name" : "limit",
-         "type" : ["number"],
-         "required" : True},
+        {"name": "comparator",
+         "type": [ "tag" ],
+         "values": [ ":over", ":under" ],
+         "required": True},
+        {"name": "limit",
+         "type": [ "number" ],
+         "required": True},
         ]
 
 class VacationCommand(ActionCommand):
     args_definition = [
-        {"name" : "subject",
-          "type" : ["tag"],
+        {"name": "days",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":subject"],
-          "extra_arg" : {"type" : "string"},
-          "required" : False},
-        {"name" : "days",
-          "type" : ["tag"],
+          "values": [ ":days" ],
+          "extra_arg": {"type": "number"},
+          "required": False},
+        {"name": "subject",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":days"],
-          "extra_arg" : {"type" : "number"},
-          "required" : False},
-        {"name" : "from",
-          "type" : ["tag"],
+          "values": [ ":subject" ],
+          "extra_arg": {"type": "string"},
+          "required": False},
+        {"name": "from",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":from"],
-          "extra_arg" : {"type" : "string"},
-          "required" : False},
-        {"name" : "addresses",
-          "type" : ["tag"],
+          "values": [ ":from" ],
+          "extra_arg": {"type": "string"},
+          "required": False},
+        {"name": "addresses",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":addresses"],
-          "extra_arg" : {"type" : "stringlist"},
-          "required" : False},
-        {"name" : "handle",
-          "type" : ["tag"],
+          "values": [ ":addresses" ],
+          "extra_arg": {"type": "stringlist"},
+          "required": False},
+        {"name": "handle",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":handle"],
-          "extra_arg" : {"type" : "string"},
-          "required" : False},
-        {"name" : "mime",
-          "type" : ["tag"],
+          "values": [ ":handle" ],
+          "extra_arg": {"type": "string"},
+          "required": False},
+        {"name": "mime",
+          "type": [ "tag" ],
           "write_tag": True,
-          "values" : [":mime"],
-          "required" : False},
-        {"name" : "reason",
-         "type" : ["string"],
-         "required" : True},
+          "values": [ ":mime" ],
+          "required": False},
+        {"name": "reason",
+         "type": [ "string" ],
+         "required": True},
         ]
 
 
