@@ -19,14 +19,17 @@ import ssl
 
 from digest_md5 import DigestMD5
 
+from binascii import b2a_base64
+
 CRLF = '\r\n'
 
-KNOWN_CAPABILITIES = ["IMPLEMENTATION", "SASL", "SIEVE", 
+KNOWN_CAPABILITIES = ["IMPLEMENTATION", "SASL", "SIEVE",
                       "STARTTLS", "NOTIFY", "LANGUAGE",
                       "RENAME"]
 
 SUPPORTED_AUTH_MECHS = ["DIGEST-MD5", "PLAIN", "LOGIN"]
 
+
 class Error(Exception):
     pass
 
@@ -85,7 +88,7 @@ class Client(object):
     def __dprint(self, message):
         if not self.__debug:
             return
-        print "DEBUG: %s" % message
+        print(message)
 
     def __read_block(self, size):
         """Read a block of 'size' bytes from the server.
@@ -112,8 +115,9 @@ class Client(object):
             return buf
         try:
             buf += self.sock.recv(size)
-        except (socket.timeout, ssl.SSLError), e:
+        except (socket.timeout, ssl.SSLError) as err:
             raise Error("Failed to read %d bytes from the server" % size)
+        self.__dprint("S: %s" % buf)
         return buf
 
     def __read_line(self):
@@ -127,7 +131,7 @@ class Client(object):
         Client.read_timeout seconds. If no data can be
         retrieved, it is considered as a fatal error and an 'Error'
         exception is raised.
-        
+
         :rtype: string
         :return: the read line
         """
@@ -160,6 +164,7 @@ class Client(object):
                 if m.group(1) == "NO":
                     self.__parse_error(m.group(2))
                 raise Response(m.group(1), m.group(2))
+        self.__dprint("S: %s" % ret)
         return ret
 
     def __read_response(self, nblines=-1):
@@ -180,11 +185,11 @@ class Client(object):
         while True:
             try:
                 line = self.__read_line()
-            except Response, inst:
+            except Response as inst:
                 code = inst.code
                 data = inst.data
                 break
-            except Literal, inst:
+            except Literal as inst:
                 resp += self.__read_block(inst.value)
                 continue
             if not len(line):
@@ -195,7 +200,7 @@ class Client(object):
                 break
 
         return (code, data, resp)
-    
+
     def __prepare_args(self, args):
         """Format command arguments before sending them.
 
@@ -233,7 +238,7 @@ class Client(object):
         tosend = name
         if len(args):
             tosend += " " + " ".join(self.__prepare_args(args))
-        self.__dprint("Command: %s" % tosend)
+        self.__dprint("C: %s" % tosend)
         self.sock.send("%s%s" % (tosend, CRLF))
         for l in extralines:
             self.sock.send("%s%s" % (l, CRLF))
@@ -242,7 +247,7 @@ class Client(object):
         if withcontent:
             return (code, data, content)
         return (code, data)
-  
+
     def __get_capabilities(self):
         code, data, capabilities = self.__read_response()
         if code == "NO":
@@ -285,15 +290,20 @@ class Client(object):
             self.errcode = ""
         self.errmsg = m.group(2).strip('"')
 
-    def _plain_authentication(self, login, password):
+    def _plain_authentication(self, login, password, authzid=None):
         """SASL PLAIN authentication
-        
+
         :param login: username
         :param password: clear password
         :return: True on success, False otherwise.
         """
-        params = base64.b64encode('\0' + '\0'.join([login, password]))
+        if not authzid == None:
+            params = b2a_base64("%s\0%s\0%s" % (authzid, login, password)).strip()
+        else:
+            params = base64.b64encode('\0' + '\0'.join([login, password]))
+
         code, data = self.__send_command("AUTHENTICATE", ["PLAIN", params])
+
         if code == "OK":
             return True
         return False
@@ -305,9 +315,9 @@ class Client(object):
         :param password: clear password
         :return: True on success, False otherwise.
         """
-        extralines = ['"%s"' % base64.b64encode(login), 
+        extralines = ['"%s"' % base64.b64encode(login),
                       '"%s"' % base64.b64encode(password)]
-        code, data = self.__send_command("AUTHENTICATE", ["LOGIN"], 
+        code, data = self.__send_command("AUTHENTICATE", ["LOGIN"],
                                          extralines=extralines)
         if code == "OK":
             return True
@@ -321,10 +331,10 @@ class Client(object):
         :return: True on success, False otherwise.
         """
         code, data, challenge = \
-            self.__send_command("AUTHENTICATE", ["DIGEST-MD5"], 
+            self.__send_command("AUTHENTICATE", ["DIGEST-MD5"],
                                 withcontent=True, nblines=1)
         dmd5 = DigestMD5(challenge, "sieve/%s" % self.srvaddr)
-        
+
         code, data, challenge = \
             self.__send_command('"%s"' % dmd5.response(login, password),
                                 withcontent=True, nblines=1)
@@ -337,7 +347,7 @@ class Client(object):
         if code == "OK":
             return True
         return False
-        
+
     def __authenticate(self, login, password, authmech=None):
         """AUTHENTICATE command
 
@@ -371,7 +381,7 @@ class Client(object):
                 self.authenticated = True
                 return True
             return False
-        
+
         self.errmsg = "No suitable mechanism found"
         return False
 
@@ -379,7 +389,7 @@ class Client(object):
         """STARTTLS command
 
         See MANAGESIEVE specifications, section 2.2.
-        
+
         :param keyfile: an eventual private key to use
         :param certfile: an eventual certificate to use
         :rtype: boolean
@@ -391,8 +401,8 @@ class Client(object):
             return False
         try:
             nsock = ssl.wrap_socket(self.sock, keyfile, certfile)
-        except ssl.SSLError, e:
-            raise Error("SSL error: %s" % str(e))
+        except ssl.SSLError as err:
+            raise Error("SSL error: %s" % str(err))
         self.sock = nsock
         self.__capabilities = {}
         self.__get_capabilities()
@@ -433,14 +443,14 @@ class Client(object):
 
         They're read from server capabilities. (see the CAPABILITY
         command)
-        
+
         :rtype: string
         """
         if type(self.__capabilities["SIEVE"]) == str:
             self.__capabilities["SIEVE"] = self.__capabilities["SIEVE"].split()
         return self.__capabilities["SIEVE"]
 
-    def connect(self, login, password, starttls=False, authmech=None):
+    def connect(self, login=None, password=None, starttls=False, authmech=None):
         """Establish a connection with the server.
 
         This function must be used. It read the server capabilities
@@ -456,16 +466,22 @@ class Client(object):
             self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
             self.sock.connect((self.srvaddr, self.srvport))
             self.sock.settimeout(Client.read_timeout)
-        except socket.error, msg:
+        except socket.error as msg:
             raise Error("Connection to server failed: %s" % str(msg))
-        
+
         if not self.__get_capabilities():
             raise Error("Failed to read capabilities from server")
+
         if starttls and not self.__starttls():
             return False
-        if self.__authenticate(login, password, authmech):
+
+        if not login == None and not password == None:
+            if self.__authenticate(login, password, authmech):
+                return True
+            else:
+                return False
+        else:
             return True
-        return False
 
     def logout(self):
         """Disconnect from the server
@@ -491,7 +507,7 @@ class Client(object):
     @authentication_required
     def havespace(self, scriptname, scriptsize):
         """Ask for available space.
-        
+
         See MANAGESIEVE specifications, section 2.5
 
         :param scriptname: script's name
@@ -564,6 +580,8 @@ class Client(object):
         content = "{%d+}%s%s" % (len(content), CRLF, content)
         code, data = \
             self.__send_command("PUTSCRIPT", [name, content])
+
+
         if code == "OK":
             return True
         return False
@@ -634,6 +652,6 @@ class Client(object):
         """
         code, data = self.__send_command("SETACTIVE", [scriptname])
         if code == "OK":
-            return True        
+            return True
         return False
 
