@@ -23,14 +23,14 @@ class ParseError(Exception):
 
 class Lexer(object):
     """
-    The lexical analysis part. 
+    The lexical analysis part.
 
     This class provides a simple way to define tokens (with patterns)
     to be detected.
 
     Patterns are provided into a list of 2-uple. Each 2-uple consists
     of a token name and an associated pattern, example:
-    
+
       [("left_bracket", r'\['),]
     """
     def __init__(self, definitions):
@@ -105,7 +105,7 @@ class Parser(object):
         if not self.debug:
             return
         for m in msgs:
-            print m
+            print(m)
 
     def __reset_parser(self):
         """Reset parser's internal variables
@@ -201,7 +201,7 @@ class Parser(object):
     def __stringlist(self, ttype, tvalue):
         """Specific method to parse the 'string-list' type
 
-        Syntax: 
+        Syntax:
             string-list = "[" string *("," string) "]" / string
                             ; if there is only a single string, the brackets
                             ; are optional
@@ -224,7 +224,7 @@ class Parser(object):
 
         This method acts as an entry point for 'argument' parsing.
 
-        Syntax: 
+        Syntax:
             string-list / number / tag
 
         :param ttype: current token type
@@ -253,7 +253,7 @@ class Parser(object):
 
         Syntax:
             *argument [ test / test-list ]
-        
+
         :param ttype: current token type
         :param tvalue: current token value
         :return: False if an error is encountered, True otherwise
@@ -307,16 +307,25 @@ class Parser(object):
 
             if ttype != "identifier":
                 return False
+
             command = get_command_instance(tvalue, self.__curcommand)
+
             if command.get_type() == "test":
                 raise ParseError("%s may not appear as a first command" % command.name)
+
             if command.get_type() == "control" and command.accept_children \
                     and command.has_arguments():
                 self.__set_expected("identifier")
+
             if self.__curcommand is not None:
-                if not self.__curcommand.addchild(command):
-                    raise ParseError("%s unexpected after a %s" % \
-                                         (tvalue, self.__curcommand.name))
+                cur_parent = self.__curcommand
+                while not cur_parent.addchild(command):
+                    if hasattr(cur_parent, 'parent'):
+                        cur_parent = cur_parent.parent
+                    else:
+                        raise ParseError("%s unexpected after a %s" % \
+                                (tvalue, self.__curcommand.name))
+
             self.__curcommand = command
             self.__cstate = self.__arguments
 
@@ -342,7 +351,7 @@ class Parser(object):
     def parse(self, text):
         """The parser entry point.
 
-        Parse the provided text to check for its validity. 
+        Parse the provided text to check for its validity.
 
         On success, the parsing tree is available into the result
         attribute. It is a list of sievecommands.Command objects (see
@@ -372,7 +381,7 @@ class Parser(object):
                                 % (ttype, "|".join(self.__expected))
                         raise ParseError(msg)
                     self.__expected = None
-                    
+
                 if not self.__command(ttype, tvalue):
                     msg = "unexpected token '%s' found near '%s'" \
                         % (tvalue, text[self.lexer.pos])
@@ -383,8 +392,8 @@ class Parser(object):
                 raise ParseError("end of script reached while %s expected" % \
                                      "|".join(self.__expected))
 
-        except (ParseError, UnknownCommand, BadArgument, BadValue), e:
-            self.error = "line %d: %s" % (self.lexer.curlineno(), str(e))
+        except (ParseError, UnknownCommand, BadArgument, BadValue) as err:
+            self.error = "line %d: %s" % (self.lexer.curlineno(), str(err))
             return False
         return True
 
@@ -421,18 +430,17 @@ if __name__ == "__main__":
     options, args = op.parse_args()
 
     if not len(args):
-        print "Nothing to parse, exiting."
+        print("Nothing to parse, exiting.")
         sys.exit(0)
 
     for a in args:
         p = Parser(debug=options.debug)
-        print "Parsing file %s... " % a,
+        print("Parsing file %s... " % (a))
         if p.parse_file(a):
-            print "OK"
+            print("OK")
             if options.verbose:
                 p.dump()
             continue
-        print "ERROR"
-        print p.error
+        print("ERROR")
+        print(p.error)
 
-    
